unit Antialiased;
//
// Displays an antialiased digonal line.
// Specify the color of the line in the forecolor
// property.  The line will be antialiased to the value
// of Color property.
//
// The module also makes 2 functions available:
//   AntialiasedLineTo, which draws an antialiased
//                      line procedurally
// and:
//   Gradiate, which creates a color gradient in an array
//
// (c) 1997-8 by Eric Fredricksen, all rights reserved
// efred@lanminds.com (tell me if you like it)
//
// Feel free to use this as you will, modify the source,
// whatever, even if you happen to be a large corporation
// with a bent towards heavy-handed anti-competitive practices.
//

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TSlope = (slUp, slDown);

  TAntialiasedLine = class(TGraphicControl)
  private
	FForeColor: TColor;
	FColors: array [0..255] of TColor;
	FDirection: TSlope;
	procedure SetForeColor(Value: TColor);
	procedure SetDirection(Value: TSlope);
	procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
  protected
	procedure Paint; override;
  public
	constructor Create(AOwner: TComponent); override;
  published
	property Color;
	property DragCursor;
	property DragMode;
	property Enabled;
	property ForeColor: TColor read FForeColor write SetForeColor;
	property ParentColor;
	property ParentShowHint;
	property ShowHint;
	property Slope: TSlope read FDirection write SetDirection;
	property Visible;
	property OnDragDrop;
	property OnDragOver;
	property OnEndDrag;
	property OnMouseDown;
	property OnMouseMove;
	property OnMouseUp;
	property OnStartDrag;
  end;

procedure AntialiasedLineTo(Canvas: TCanvas; X, Y: Integer);
  // easy access to antialiasing via simple function
  // works like LineTo but Draws an antialiased line
  // from the current pen position to X,Y in the current
  // pen color with the current brush color as the background color

procedure Gradiate(Begindex, Endex: Integer; var Rainbo: array of TColor);
  // bonus function: create a color gradient in Rainbo starting with
  // the color at Begindex and to the color at Endex

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Home Brewed', [TAntialiasedLine]);
end;

procedure AntialiasedLine(c: TCanvas; x0, y0, x1, y1: Integer; caler: array of TColor);
const
  BaseColor = 0;
  NumLevels = 256;
  IntensityBits = 8;
var
  IntensityShift, ErrorAdj, ErrorAcc: Word;
  ErrorAccTemp, Weighting, WeightingComplementMask: Word;
  DeltaX, DeltaY, Temp, XDir: Integer;
begin
  if y0 > y1 then begin
    Temp := y0; y0 := y1; y1 := Temp;
	Temp := x0; x0 := x1; x1 := Temp;
  end;

  c.Pixels[x0,y0] := Caler[BaseColor];

  DeltaX := x1-x0;

  if DeltaX >= 0 then
    XDir := 1
  else begin
	XDir := -1;
    DeltaX := -DeltaX;
  end;

  DeltaY := y1 - y0;

  // special case some lineses...

  if deltaY = 0 then begin
    // hoizontal
	while deltaX <> 0 do begin
      Dec(DeltaX);
      Inc(x0,xDir);
      c.Pixels[x0,y0] := Caler[BaseColor];
    end;
    Exit;
  end;

  if deltaX = 0 then begin
    // vertical
	repeat
      Inc(y0);
      c.Pixels[x0,y0] := Caler[BaseColor];
      Dec(deltaY);
    until deltaY = 0;
    Exit;
  end;

  if deltaX = deltaY then begin
    // diagonal
	repeat
      Inc(x0,xDir);
      Inc(y0);
      c.Pixels[x0,y0] := Caler[BaseColor];
      Dec(deltaY);
    until deltaY = 0;
	Exit;
  end;

  // normal odd cases

  ErrorAcc := 0;

  IntensityShift := 16 - IntensityBits;

  WeightingComplementMask := NumLevels - 1;

  if deltaY > deltaX then begin
    // y-maj line
    ErrorAdj := (deltaX shl 16) div deltaY;

    while deltaY > 1 do begin
      Dec(deltaY);
      ErrorAccTemp := ErrorAcc;
      Inc(ErrorAcc, ErrorAdj);
      if ErrorAcc <= ErrorAccTemp then
		Inc(x0,XDir);

      Inc(y0);

	  Weighting := ErrorAcc shr IntensityShift;

      c.Pixels[x0,     y0] := Caler[BaseColor + Weighting];
      c.Pixels[x0+XDir,y0] := Caler[BaseColor + (Weighting xor WeightingComplementMask)];
    end;

  end else begin

    // x-major

	ErrorAdj := (deltaY shl 16) div deltaX;

    while deltaX > 1 do begin
      Dec(deltaX);
      ErrorAccTemp := ErrorAcc;
      Inc(ErrorAcc, ErrorAdj);
	  if ErrorAcc <= ErrorAccTemp then
        Inc(y0);

      Inc(x0,XDir);

      Weighting := ErrorAcc shr IntensityShift;

      c.Pixels[x0,  y0] := Caler[BaseColor + Weighting];
      c.Pixels[x0,y0+1] := Caler[BaseColor + Weighting xor WeightingComplementMask];
    end;

  end;

  // final pixel
  c.Pixels[x1,y1] := Caler[BaseColor];
end;


procedure ThickAntialiasedLine(th: Integer; c: TCanvas; x0, y0, x1, y1: Integer; caler: array of TColor);
const
  BaseColor = 0;
  NumLevels = 256;
  IntensityBits = 8;
var
  IntensityShift, ErrorAdj, ErrorAcc: Word;
  ErrorAccTemp, Weighting, WeightingComplementMask: Word;
  DeltaX, DeltaY, Temp, XDir: Integer;
  i: Integer;
begin
  if y0 > y1 then begin
    Temp := y0; y0 := y1; y1 := Temp;
    Temp := x0; x0 := x1; x1 := Temp;
  end;

  DeltaX := x1-x0;

  if DeltaX >= 0 then
    XDir := 1
  else begin
    XDir := -1;
	DeltaX := -DeltaX;
  end;

  DeltaY := y1 - y0;


  ErrorAcc := 0;

  IntensityShift := 16 - IntensityBits;

  WeightingComplementMask := NumLevels - 1;

  if deltaY > deltaX then begin

	// y-maj line

    if XDir > 0 then begin
      Dec( x0, th div 2 );
      for i := x0 to x0+th do
        c.Pixels[i,y0] := Caler[BaseColor];
	end else begin
      Inc( x0, th div 2 );
      for i := x0 downto x0-th do
        c.Pixels[i,y0] := Caler[BaseColor];
	end;

    ErrorAdj := (deltaX shl 16) div deltaY;

    while deltaY > 0 do begin
      Dec(deltaY);
      ErrorAccTemp := ErrorAcc;
      Inc(ErrorAcc, ErrorAdj);
      if ErrorAcc <= ErrorAccTemp then
        Inc(x0,XDir);

      Inc(y0);

      Weighting := ErrorAcc shr IntensityShift;

      c.Pixels[x0,     y0] := Caler[BaseColor + Weighting];
	  if XDir > 0 then
        for i := x0+1 to x0+th do
          c.Pixels[i,y0] := Caler[BaseColor]
      else
		for i := x0-1 downto x0-th do
          c.Pixels[i,y0] := Caler[BaseColor];
      c.Pixels[i,y0] := Caler[BaseColor + (Weighting xor WeightingComplementMask)];
    end;

  end else begin

    // x-major

	Dec( x0, th div 2 );
    for i := y0 to y0+th do
      c.Pixels[x0,i] := Caler[BaseColor];

    ErrorAdj := (deltaY shl 16) div deltaX;

    while deltaX > 0 do begin
      Dec(deltaX);
      ErrorAccTemp := ErrorAcc;
      Inc(ErrorAcc, ErrorAdj);
      if ErrorAcc <= ErrorAccTemp then
		Inc(y0);

      Inc(x0,XDir);

      Weighting := ErrorAcc shr IntensityShift;

	  c.Pixels[x0,  y0] := Caler[BaseColor + Weighting];
	  for i := y0+1 to y0+th do
		c.Pixels[x0,i] := Caler[BaseColor];
	  c.Pixels[x0,i] := Caler[BaseColor + Weighting xor WeightingComplementMask];
	end;

  end;

end;

procedure Gradiate(Begindex, Endex: Integer; var Rainbo: array of TColor);
var
  i: Integer;
  reds, greens, blues, steps: Integer;
  peRed, peGreen, peBlue: Integer;
  begR, begG, begB: Integer;
begin
  if (Begindex = Endex) or (Begindex < 0) or (Endex < 0) then Exit;

  begR := Rainbo[Begindex] and $FF;
  begG := (Rainbo[Begindex] and $FF00) shr 8;
  begB := (Rainbo[Begindex] and $FF0000) shr 16;

  reds :=   (Rainbo[Endex] and $FF) -
            begR;
  greens := ((Rainbo[Endex] and $FF00) shr 8) -
			begG;
  blues  := ((Rainbo[Endex] and $FF0000) shr 16) -
            begB;
  steps := Endex - Begindex;

  for i := Begindex + 1 to Endex - 1 do
  begin
	peRed   := begR + (reds   * (i-Begindex)) div steps;
	peGreen := begG + (greens * (i-Begindex)) div steps;
    peBlue  := begB  + (blues  * (i-Begindex)) div steps;
    Rainbo[i] := peRed + (peGreen shl 8) + (peBlue shl 16);
  end;
end;

procedure AntialiasedLineTo(Canvas: TCanvas; X, Y: Integer);
var
  caler: array[0..255] of TColor;
begin
  caler[0] := Canvas.Pen.Color;
  caler[255] := Canvas.Brush.Color;
  Gradiate(0,255,caler);

  if Canvas.Pen.Width > 1 then
    ThickAntialiasedLine( Canvas.Pen.Width,
                     Canvas,
                     Canvas.PenPos.X, Canvas.PenPos.Y,
                     X, Y, caler )
  else
	AntialiasedLine( Canvas,
                     Canvas.PenPos.X, Canvas.PenPos.Y,
                     X, Y, caler );

  Canvas.MoveTo(X,Y);
end;

{ TAnialiasedLine }

constructor TAntialiasedLine.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable];
  Width := 65;
  Height := 45;
  FForeColor := clBlack;

  FDirection := slDown;

  FColors[0] := ColorToRGB(FForeColor);
  FColors[255] := ColorToRGB(Color);
  Gradiate(0,255,FColors);
end;

procedure TAntialiasedLine.Paint;
begin
  if FDirection = slDown then
    AntialiasedLine(Canvas,0,0,Width-1,Height-1,FColors)
  else
    AntialiasedLine(Canvas,0,Height-1,Width-1,0,FColors);
end;

procedure TAntialiasedLine.SetDirection(Value: TSlope);
begin
  if FDirection <> Value then
  begin
    FDirection := Value;
    Invalidate;
  end;
end;

procedure TAntialiasedLine.SetForeColor(Value: TColor);
begin
  if FForeColor <> Value then
  begin
	FForeColor := Value;
    FColors[0] := ColorToRGB(FForeColor);
    Gradiate(0,255,FColors);
    Invalidate;
  end;
end;

procedure TAntialiasedLine.CMColorChanged(var Message: TMessage);
begin
  FColors[255] := ColorToRGB(Color);
  Gradiate(0,255,FColors);
  Invalidate;
end;

end.

